"""Services for the job_script_templates resource, including module specific business logic."""

from typing import Any

from buzz import require_condition
from fastapi import status
from loguru import logger
from pendulum.datetime import DateTime as PendulumDateTime
from sqlalchemy import delete, func, not_, select, update
from sqlalchemy.sql.expression import Select

from jobbergate_api.apps.job_script_templates.models import JobScriptTemplate
from jobbergate_api.apps.job_scripts.models import JobScript
from jobbergate_api.apps.services import AutoCleanResponse, CrudModel, CrudService, FileService, ServiceError
from jobbergate_api.config import settings


class JobScriptTemplateService(CrudService):
    """
    Provide a CrudService that overloads the list query builder and locator logic.
    """

    def build_list_query(
        self,
        sort_ascending: bool = True,
        search: str | None = None,
        sort_field: str | None = None,
        include_archived: bool = True,
        include_files: bool = False,
        include_parent: bool = False,
        include_null_identifier: bool = True,
        **additional_filters,
    ) -> Select:
        """List all job_script_templates."""
        query: Select = super().build_list_query(
            sort_ascending=sort_ascending,
            search=search,
            sort_field=sort_field,
            include_archived=include_archived,
            include_files=include_files,
            include_parent=include_parent,
            **additional_filters,
        )
        if not include_null_identifier:
            query = query.where(not_(JobScriptTemplate.identifier.is_(None)))
        return query

    async def create(self, **incoming_data) -> CrudModel:
        """
        Add a new row for the model to the database.
        """
        self.validate_identifier(incoming_data.get("identifier"))
        return await super().create(**incoming_data)

    async def update(self, locator: Any, **incoming_data) -> CrudModel:
        """
        Update a row by locator with supplied data.
        """
        self.validate_identifier(incoming_data.get("identifier"))
        return await super().update(locator, **incoming_data)

    def locate_where_clause(self, id_or_identifier: Any) -> Any:
        """
        Locate an instance using the id or identifier field.
        """
        if isinstance(id_or_identifier, str):
            return JobScriptTemplate.identifier == id_or_identifier
        elif isinstance(id_or_identifier, int):
            return JobScriptTemplate.id == id_or_identifier
        else:
            raise ValueError("id_or_identifier must be a string or integer")

    def validate_identifier(self, identifier: str | None) -> None:
        """
        Validate that the identifier is not an empty string nor composed only by digits.

        Raise a ServiceError with status code 422 if the validation fails.

        Many of the job-script-template endpoints use the id or identifier interchangeably
        as a path parameter. With that, we need to ensure that the identifier is
        not a number, as that would be identified as id.
        """
        if identifier is not None:
            require_condition(
                identifier.strip() and not identifier.isdigit(),
                "Identifier on {} can not be a empty string nor be composed only by digits. Got {}".format(
                    self.name, identifier
                ),
                raise_exc_class=ServiceError,
                raise_kwargs=dict(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY),
            )

    async def clean_unused_entries(self) -> AutoCleanResponse:
        """
        Automatically clean unused job script templates depending on a threshold.

        Based on the last time each job script template was updated or used to create a job script,
        this will archive job script templates that were unarchived and delete job script templates
        that were archived.
        """
        result = AutoCleanResponse(archived=set(), deleted=set())

        if settings.AUTO_CLEAN_JOB_SCRIPT_TEMPLATES_DAYS_TO_DELETE is not None:
            days_to_delete = settings.AUTO_CLEAN_JOB_SCRIPT_TEMPLATES_DAYS_TO_DELETE
            threshold = PendulumDateTime.utcnow().subtract(days=days_to_delete).naive()
            subquery_unused_job_script_templates = select(self.model_type.id).where(
                self.model_type.is_archived.is_(True), self.model_type.updated_at < threshold
            )
            subquery_recent_child = (
                select(JobScript.parent_template_id)
                .where(JobScript.parent_template_id.in_(subquery_unused_job_script_templates))
                .group_by(JobScript.parent_template_id)
                .having(func.max(JobScript.created_at) >= threshold)
            )

            delete_query = (
                delete(self.model_type)  # type: ignore
                .where(
                    self.model_type.id.in_(subquery_unused_job_script_templates),
                    ~self.model_type.id.in_(subquery_recent_child),
                )
                .returning(self.model_type.id)
            )
            deleted = await self.session.execute(delete_query)
            result.deleted.update(row[0] for row in deleted.all())
        logger.debug(f"Job script templates deleted: {result.deleted}")

        if settings.AUTO_CLEAN_JOB_SCRIPT_TEMPLATES_DAYS_TO_ARCHIVE is not None:
            days_to_archive = settings.AUTO_CLEAN_JOB_SCRIPT_TEMPLATES_DAYS_TO_ARCHIVE
            threshold = PendulumDateTime.utcnow().subtract(days=days_to_archive).naive()
            subquery_unused_job_script_templates = select(self.model_type.id).where(
                self.model_type.is_archived.is_(False), self.model_type.updated_at < threshold
            )
            subquery_recent_child = (
                select(JobScript.parent_template_id)
                .where(JobScript.parent_template_id.in_(subquery_unused_job_script_templates))
                .group_by(JobScript.parent_template_id)
                .having(func.max(JobScript.created_at) >= threshold)
            )

            update_query = (
                update(self.model_type)  # type: ignore
                .where(
                    self.model_type.id.in_(subquery_unused_job_script_templates),
                    ~self.model_type.id.in_(subquery_recent_child),
                )
                .values(is_archived=True)
                .returning(self.model_type.id)
            )
            archived = await self.session.execute(update_query)
            result.archived.update(row[0] for row in archived.all())
        logger.debug(f"Job script templates marked as archived: {result.archived}")

        return result


class JobScriptTemplateFileService(FileService):
    """
    Provide an empty derived class of FileService.

    Although it doesn't do anything, it fixes errors with mypy:
        error: Value of type variable "FileModel" of "FileService" cannot be "JobScriptTemplateFile"
        error: Value of type variable "FileModel" of "FileService" cannot be "WorkflowFile"
    """



"""Provide protocols for CRUD and file operations in routers in a dedicated file to avoid circular imports."""

from __future__ import annotations

from typing import Protocol, TypeVar

from sqlalchemy.orm import Mapped
from sqlalchemy.sql.expression import Select


class CrudModelProto(Protocol):
    """
    Provide a protocol for models that can be operated on by the CrudService.

    This protocol enables type hints for editors and type checking with mypy.

    These services would best be served by an intersection type so that the model_type is actually
    specified to inherit from _both_ the mixins and the Base. This would allow static type checkers to
    recognize that all of the columns in a mixin are available and that the class can be
    instantiated in the create method. However, intersection types are not supported yet. For more
    information, see this discussion: https://github.com/python/typing/issues/213
    """

    id: Mapped[int]
    owner_email: Mapped[str]
    is_archived: Mapped[bool]

    def __init__(self, **kwargs):
        """
        Declare that the protocol can be instantiated.
        """
        ...

    def __tablename__(self) -> str:
        """
        Declare that the protocol has a method to dynamically produce the table name.
        """
        ...

    @classmethod
    def searchable_fields(cls) -> set[str]:
        """
        Declare that the protocol has searchable fields.
        """
        ...

    @classmethod
    def sortable_fields(cls) -> set[str]:
        """
        Declare that the protocol has sortable fields.
        """
        ...

    @classmethod
    def include_files(cls, query: Select) -> Select:
        """
        Declare that the protocol has a method to include files in a query.
        """
        ...

    @classmethod
    def include_parent(cls, query: Select) -> Select:
        """
        Declare that the protocol has a method to include details about the parent entry in a query.
        """
        ...


CrudModel = TypeVar("CrudModel", bound=CrudModelProto)


class FileModelProto(Protocol):
    """
    Provide a protocol for models that can be operated on by the FileService.

    This protocol enables type hints for editors and type checking with mypy.

    These services would best be served by an intersection type so that the model_type is actually
    specified to inherit from _both_ the mixins and the Base. This would allow static type checkers to
    recognize that all of the columns in a mixin are available and that the class can be
    instantiated in the create method. However, intersection types are not supported yet. For more
    information, see this discussion: https://github.com/python/typing/issues/213
    """

    parent_id: Mapped[int]
    filename: Mapped[str]
    file_key: str

    def __init__(self, **kwargs):
        """
        Declare that the protocol can be instantiated.
        """
        ...

    def __tablename__(self) -> str:
        """
        Declare that the protocol has a method to dynamically produce the table name.
        """
        ...


FileModel = TypeVar("FileModel", bound=FileModelProto)



"""Core helper functions for job submissions."""

from collections.abc import Iterable
from math import ceil
from textwrap import dedent
from typing import Any, assert_never, Type

from loguru import logger

from jobbergate_api.apps.job_submissions.constants import (
    JobSubmissionMetricSampleRate,
    JobSubmissionMetricAggregateNames,
)


def validate_job_metric_upload_input(
    data: Any, expected_types: tuple[Type[Any], ...]
) -> Iterable[tuple[Any]]:
    """Validate if the input data of job metric upload is correct once decoded.

    It will brute force apply the expected types to the data and raise an error in case it fails.

    Args:
        data (Iterable[list[Any] | tuple[Any]]): The decoded data, which should be a list of lists or tuples,
            where each inner list or tuple contains the data for a single job metric upload.
        expected_types (tuple[Type[Any], ...]): A tuple of types that each element in the inner lists or tuples
            should match.

    Returns:
        Iterable[list[Any] | tuple[Any]]: The validated data.
    """

    def _force_cast(object: Any, expected_type: Type[Any]) -> Any:
        return expected_type(object)

    if not isinstance(data, list):
        raise ValueError("Decoded data must be a list.")
    if not all(isinstance(x, (list, tuple)) for x in data):
        raise ValueError("All elements of the inner data must be a Sequence")
    if not all(len(x) == len(expected_types) for x in data):
        raise ValueError("Every iterable in `data` must match the length of `expected_types`.")
    # postgres limits the number of query params to 2**15 - 1
    # https://www.postgresql.org/docs/17/protocol-message-formats.html
    if len(expected_types) * len(data) > 2**15 - 1:
        raise ValueError(
            "The maximum number of elements has been exceeded. " "Maximum is {}, received {}".format(
                ceil((2**15 - 1) / len(expected_types)), len(data)
            )
        )
    for idx, aggregated_data in enumerate(data):
        # for each element in the inner list or tuple, force apply the expected type
        try:
            data[idx] = tuple(
                _force_cast(data, expected_types[idx]) for idx, data in enumerate(aggregated_data)
            )
        except Exception as e:
            logger.error(f"Failed to cast data to expected types: {e}")
            raise ValueError("Failed to cast data to expected types.")
    return data


def build_job_metric_aggregation_query(node: str | None, sample_rate: JobSubmissionMetricSampleRate) -> str:
    """
    Build a SQL query string to aggregate job metrics based on the provided node and sample rate.

    Args:
        node (str | None): The node host identifier. If None, the query will aggregate metrics for all nodes.
        sample_rate (JobSubmissionMetricSampleRate): The sample rate for the metrics aggregation. Determines the view name to use.

    Returns:
        str: The SQL query string for aggregating job metrics.
    """
    if node is not None:
        where_statement = "WHERE job_submission_id = :job_submission_id AND node_host = :node_host"
        match sample_rate:
            case JobSubmissionMetricSampleRate.ten_seconds:
                view_name = JobSubmissionMetricAggregateNames.metrics_nodes_mv_10_seconds_by_node
            case JobSubmissionMetricSampleRate.one_minute:
                view_name = JobSubmissionMetricAggregateNames.metrics_nodes_mv_1_minute_by_node
            case JobSubmissionMetricSampleRate.ten_minutes:
                view_name = JobSubmissionMetricAggregateNames.metrics_nodes_mv_10_minutes_by_node
            case JobSubmissionMetricSampleRate.one_hour:
                view_name = JobSubmissionMetricAggregateNames.metrics_nodes_mv_1_hour_by_node
            case JobSubmissionMetricSampleRate.one_week:
                view_name = JobSubmissionMetricAggregateNames.metrics_nodes_mv_1_week_by_node
            case _ as unreachable:
                assert_never(unreachable)
    else:
        where_statement = "WHERE job_submission_id = :job_submission_id"
        match sample_rate:
            case JobSubmissionMetricSampleRate.ten_seconds:
                view_name = JobSubmissionMetricAggregateNames.metrics_nodes_mv_10_seconds_all_nodes
            case JobSubmissionMetricSampleRate.one_minute:
                view_name = JobSubmissionMetricAggregateNames.metrics_nodes_mv_1_minute_all_nodes
            case JobSubmissionMetricSampleRate.ten_minutes:
                view_name = JobSubmissionMetricAggregateNames.metrics_nodes_mv_10_minutes_all_nodes
            case JobSubmissionMetricSampleRate.one_hour:
                view_name = JobSubmissionMetricAggregateNames.metrics_nodes_mv_1_hour_all_nodes
            case JobSubmissionMetricSampleRate.one_week:
                view_name = JobSubmissionMetricAggregateNames.metrics_nodes_mv_1_week_all_nodes
            case _ as unreachable:
                assert_never(unreachable)

    return dedent(
        f"""
        SELECT bucket,
            node_host,
            cpu_frequency,
            cpu_time,
            cpu_utilization,
            gpu_memory,
            gpu_utilization,
            page_faults,
            memory_rss,
            memory_virtual,
            disk_read,
            disk_write
        FROM {view_name}
        {where_statement}
        AND bucket >= :start_time
        AND bucket <= :end_time
        ORDER BY bucket
        """
    )



"""Delete unused files from Jobbergate's file storage."""

import asyncio
from dataclasses import dataclass

from loguru import logger
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from jobbergate_api.apps.protocols import FileModelProto
from jobbergate_api.safe_types import Bucket


@dataclass
class GarbageCollector:
    """
    Class to delete unused files from Jobbergate's file storage.
    """

    model_type: type[FileModelProto]
    bucket: Bucket
    session: AsyncSession

    def __post_init__(self) -> None:
        """Post-initialize method to set up extra internal details."""
        MAX_CONCURRENT_REQUESTS = 25
        self.semaphore = asyncio.Semaphore(MAX_CONCURRENT_REQUESTS)

    async def run(self) -> None:
        """Delete files from the bucket."""
        logger.debug(f"Running garbage collector for {self.model_type.__tablename__}")
        files_to_delete = await self._get_files_to_delete()
        tasks = [self._delete_file(file) for file in files_to_delete]
        await asyncio.gather(*tasks)

    async def _delete_file(self, file: str) -> None:
        async with self.semaphore:
            obj = await self.bucket.Object(file)
            await obj.delete()
        logger.debug(f"Deleted file {file} from bucket {self.bucket.name}")

    async def _get_set_of_files_from_database(self) -> set[str]:
        """Get a set of files from the database."""
        rows = await self.session.execute(select(self.model_type))  # type: ignore
        result = {obj.file_key for obj in rows.scalars().all()}
        logger.debug(f"Total of files found in the table {self.model_type.__tablename__}: {len(result)}")
        return result

    async def _get_set_of_files_from_bucket(self) -> set[str]:
        """Get a set of files from the bucket."""
        prefix = self.model_type.__tablename__
        result = {obj.key async for obj in self.bucket.objects.filter(Prefix=prefix)}
        logger.debug(
            f"Total of files found in the bucket {self.bucket.name} with prefix {prefix}: {len(result)}"
        )
        return result

    async def _get_files_to_delete(self) -> set[str]:
        """Get a set of files to delete."""
        files_in_database = await self._get_set_of_files_from_database()
        files_in_bucket = await self._get_set_of_files_from_bucket()
        result = files_in_bucket - files_in_database
        logger.debug(f"Total of files to be garbage collected: {len(result)}")
        return result



"""Services for the job_submissions resource, including module specific business logic."""

from loguru import logger
from pendulum.datetime import DateTime as PendulumDateTime
from sqlalchemy import delete, select, update
from sqlalchemy.sql.expression import Select

from jobbergate_api.apps.job_submissions.models import JobSubmission
from jobbergate_api.apps.services import AutoCleanResponse, CrudService
from jobbergate_api.config import settings


class JobSubmissionService(CrudService):
    """
    Provide a CrudService that overloads the list query builder.
    """

    def build_list_query(
        self,
        sort_ascending: bool = True,
        search: str | None = None,
        sort_field: str | None = None,
        include_archived: bool = True,
        include_files: bool = False,
        include_parent: bool = False,
        filter_slurm_job_ids: list[int] | None = None,
        **additional_filters,
    ) -> Select:
        """
        List all job_script_templates.
        """
        query: Select = super().build_list_query(
            sort_ascending=sort_ascending,
            search=search,
            sort_field=sort_field,
            include_archived=include_archived,
            include_files=include_files,
            include_parent=include_parent,
            **additional_filters,
        )
        if filter_slurm_job_ids:
            query = query.where(JobSubmission.slurm_job_id.in_(filter_slurm_job_ids))
        return query

    async def clean_unused_entries(self) -> AutoCleanResponse:
        """
        Automatically clean unused job submissions depending on a threshold.

        Based on the last time each job submission was updated, this will archive job
        submissions that were unarchived and delete job submissions that were archived.
        """
        result = AutoCleanResponse(archived=set(), deleted=set())

        if settings.AUTO_CLEAN_JOB_SUBMISSIONS_DAYS_TO_DELETE is not None:
            days_to_delete = settings.AUTO_CLEAN_JOB_SUBMISSIONS_DAYS_TO_DELETE
            threshold = PendulumDateTime.utcnow().subtract(days=days_to_delete).naive()
            query_unused_job_submissions = select(self.model_type.id).where(
                self.model_type.is_archived.is_(True), self.model_type.updated_at < threshold
            )

            delete_query = (
                delete(self.model_type)  # type: ignore
                .where(self.model_type.id.in_(query_unused_job_submissions))
                .returning(self.model_type.id)
            )
            deleted = await self.session.execute(delete_query)
            result.deleted.update(row[0] for row in deleted.all())
        logger.debug(f"Job submissions deleted: {result.deleted}")
        logger.debug(f"{settings.AUTO_CLEAN_JOB_SUBMISSIONS_DAYS_TO_ARCHIVE=}")

        if settings.AUTO_CLEAN_JOB_SUBMISSIONS_DAYS_TO_ARCHIVE is not None:
            days_to_archive = settings.AUTO_CLEAN_JOB_SUBMISSIONS_DAYS_TO_ARCHIVE
            threshold = PendulumDateTime.utcnow().subtract(days=days_to_archive).naive()
            query_unused_job_submissions = select(self.model_type.id).where(
                self.model_type.is_archived.is_(False), self.model_type.updated_at < threshold
            )

            update_query = (
                update(self.model_type)  # type: ignore
                .where(self.model_type.id.in_(query_unused_job_submissions))
                .values(is_archived=True)
                .returning(self.model_type.id)
            )
            archived = await self.session.execute(update_query)
            result.archived.update(row[0] for row in archived.all())
        logger.debug(f"Job submissions marked as archived: {result.archived}")

        return result


class JobProgressService(CrudService):
    """
    Provide a CrudService that overloads the list query builder.
    """

    def build_list_query(
        self,
        sort_ascending: bool = True,
        search: str | None = None,
        sort_field: str | None = None,
        include_archived: bool = True,
        include_files: bool = False,
        include_parent: bool = False,
        **additional_filters,
    ) -> Select:
        """
        List all entries in the job_progress table for a given job_submission_id.
        """
        query: Select = super().build_list_query(
            sort_ascending=sort_ascending,
            sort_field=sort_field,
            **additional_filters,
        )
        return query



"""
JobScript resource schema.
"""

from datetime import datetime
from textwrap import dedent
from typing import Any

from pydantic import BaseModel, ConfigDict, NonNegativeInt

from jobbergate_api.apps.constants import FileType
from jobbergate_api.apps.job_script_templates.schemas import JobTemplateListView
from jobbergate_api.apps.schemas import LengthLimitedStr, TableResource
from jobbergate_api.meta_mapper import MetaField, MetaMapper

job_script_meta_mapper = MetaMapper(
    id=MetaField(
        description="The unique database identifier for the instance",
        example=101,
    ),
    parent_id=MetaField(
        description="The unique database identifier for the parent of this instance",
        example=101,
    ),
    created_at=MetaField(
        description="The timestamp for when the instance was created",
        example="2023-08-18T13:55:37.172285",
    ),
    updated_at=MetaField(
        description="The timestamp for when the instance was last updated",
        example="2023-08-18T13:55:37.172285",
    ),
    name=MetaField(
        description="The unique name of the instance",
        example="test-job-script-88",
    ),
    description=MetaField(
        description="A text field providing a human-friendly description of the job_script",
        example="This job_scripts runs an Foo job using the bar variant",
    ),
    owner_email=MetaField(
        description="The email of the owner/creator of the instance",
        example="tucker@omnivector.solutions",
    ),
    parent_template_id=MetaField(
        description="The foreign-key to the job script template from which this instance was created",
        example=71,
    ),
    sbatch_params=MetaField(
        description="SBATCH parameters to inject into the job_script",
        example=["alpha", "beta", "delta"],
    ),
    param_dict=MetaField(
        description="Parameters to use when rendering the job_script jinja2 template",
        example={"param1": 7, "param2": 13},
    ),
    template_output_name_mapping=MetaField(
        description=dedent(
            """
            A mapping of template names to file names.
            The first element is the entrypoint, the others are optional support files.
            """
        ).strip(),
        example={"template.py.jinja2": "template.py", "support.json.jinja2": "support.json"},
    ),
    filename=MetaField(
        description="The name of the file",
        example="job-script.py",
    ),
    file_type=MetaField(
        description="The type of the file",
        example=FileType.ENTRYPOINT.value,
    ),
    is_archived=MetaField(
        description="Indicates if the job script has been archived.",
        example=False,
    ),
    cloned_from_id=MetaField(
        description="Indicates the id this entry has been cloned from, if any.",
        example=101,
    ),
)


class JobScriptCreateRequest(BaseModel):
    """
    Request model for creating JobScript instances.
    """

    name: LengthLimitedStr
    description: LengthLimitedStr | None = None

    model_config = ConfigDict(json_schema_extra=job_script_meta_mapper)


class JobScriptCloneRequest(BaseModel):
    """
    Request model for cloning JobScript instances.
    """

    name: LengthLimitedStr | None = None
    description: LengthLimitedStr | None = None

    model_config = ConfigDict(json_schema_extra=job_script_meta_mapper)


class RenderFromTemplateRequest(BaseModel):
    """Request model for creating a JobScript entry from a template."""

    template_output_name_mapping: dict[LengthLimitedStr, LengthLimitedStr]
    sbatch_params: list[str] | None = None
    param_dict: dict[str, Any]

    model_config = ConfigDict(json_schema_extra=job_script_meta_mapper)


class JobScriptUpdateRequest(BaseModel):
    """
    Request model for updating JobScript instances.
    """

    name: LengthLimitedStr | None = None
    description: LengthLimitedStr | None = None
    is_archived: bool | None = None

    model_config = ConfigDict(json_schema_extra=job_script_meta_mapper)


class JobScriptFileDetailedView(BaseModel):
    """Model for the job_script_files field of the JobScript resource."""

    parent_id: NonNegativeInt
    filename: LengthLimitedStr
    file_type: FileType
    created_at: datetime | None = None
    updated_at: datetime | None = None

    model_config = ConfigDict(from_attributes=True, json_schema_extra=job_script_meta_mapper)


class JobScriptBaseView(TableResource):
    """
    Base model to match database for the JobScript resource.

    Omits parent relationship.
    """

    parent_template_id: int | None = None
    cloned_from_id: int | None = None

    model_config = ConfigDict(json_schema_extra=job_script_meta_mapper)


class JobScriptListView(JobScriptBaseView):
    """Model to match database for the JobScript resource."""

    template: JobTemplateListView | None = None


class JobScriptDetailedView(JobScriptBaseView):
    """Model to match database for the JobScript resource."""

    files: list[JobScriptFileDetailedView] | None



"""Services for the job_scripts resource, including module specific business logic."""

from typing import Any

from buzz import enforce_defined, require_condition
from fastapi import UploadFile, status
from loguru import logger
from pendulum.datetime import DateTime as PendulumDateTime
from pydantic import AnyUrl
from sqlalchemy import delete, func, select, update

from jobbergate_api.apps.constants import FileType
from jobbergate_api.apps.job_submissions.constants import JobSubmissionStatus
from jobbergate_api.apps.job_submissions.models import JobSubmission
from jobbergate_api.apps.services import AutoCleanResponse, CrudService, FileModel, FileService, ServiceError
from jobbergate_api.config import settings


class JobScriptCrudService(CrudService):
    """
    Provide an empty derived class of CrudService.

    Although it doesn't do anything, it fixes an error with mypy:
        error: Value of type variable "CrudModel" of "CrudService" cannot be "JobScript"
    """

    async def delete(self, locator: Any) -> None:
        """
        Extend delete a row by locator.

        Orphaned job-scripts are now allowed on Jobbergate. However, the agent
        relies on them to submit jobs after requesting GET /agent/pending.
        This creates a race condition and errors occur when a job-script is
        deleted before the agent handles its submissions.

        To avoid this, they are marked as reject in this scenario.
        """
        query = (
            update(JobSubmission)  # type: ignore
            .where(JobSubmission.job_script_id == locator)
            .where(JobSubmission.status == JobSubmissionStatus.CREATED)
            .values(
                status=JobSubmissionStatus.REJECTED,
                report_message="Parent job script was deleted before the submission.",
            )
        )
        await self.session.execute(query)
        await super().delete(locator)

    async def clean_unused_entries(self) -> AutoCleanResponse:
        """
        Automatically clean unused job scripts depending on a threshold.

        Based on the last time each job script was updated or used to create a job submission,
        this will archive job scripts that were unarchived and delete job scripts that were archived.
        """
        result = AutoCleanResponse(archived=set(), deleted=set())

        if settings.AUTO_CLEAN_JOB_SCRIPTS_DAYS_TO_DELETE is not None:
            days_to_delete = settings.AUTO_CLEAN_JOB_SCRIPTS_DAYS_TO_DELETE
            threshold = PendulumDateTime.utcnow().subtract(days=days_to_delete).naive()
            subquery_unused_job_script = select(self.model_type.id).where(
                self.model_type.is_archived.is_(True), self.model_type.updated_at < threshold
            )
            subquery_recent_child = (
                select(JobSubmission.job_script_id)
                .where(JobSubmission.job_script_id.in_(subquery_unused_job_script))
                .group_by(JobSubmission.job_script_id)
                .having(func.max(JobSubmission.created_at) >= threshold)
            )

            delete_query = (
                delete(self.model_type)  # type: ignore
                .where(
                    self.model_type.id.in_(subquery_unused_job_script),
                    ~self.model_type.id.in_(subquery_recent_child),
                )
                .returning(self.model_type.id)
            )
            deleted = await self.session.execute(delete_query)
            result.deleted.update(row[0] for row in deleted.all())
        logger.debug(f"Job scripts deleted: {result.deleted}")

        if settings.AUTO_CLEAN_JOB_SCRIPTS_DAYS_TO_ARCHIVE is not None:
            days_to_archive = settings.AUTO_CLEAN_JOB_SCRIPTS_DAYS_TO_ARCHIVE
            threshold = PendulumDateTime.utcnow().subtract(days=days_to_archive).naive()
            subquery_unused_job_script = select(self.model_type.id).where(
                self.model_type.is_archived.is_(False), self.model_type.updated_at < threshold
            )
            subquery_recent_child = (
                select(JobSubmission.job_script_id)
                .where(JobSubmission.job_script_id.in_(subquery_unused_job_script))
                .group_by(JobSubmission.job_script_id)
                .having(func.max(JobSubmission.created_at) >= threshold)
            )

            update_query = (
                update(self.model_type)  # type: ignore
                .where(
                    self.model_type.id.in_(subquery_unused_job_script),
                    ~self.model_type.id.in_(subquery_recent_child),
                )
                .values(is_archived=True)
                .returning(self.model_type.id)
            )
            archived = await self.session.execute(update_query)
            result.archived.update(row[0] for row in archived.all())
        logger.debug(f"Job scripts marked as archived: {result.archived}")

        return result


class JobScriptFileService(FileService):
    """
    Provide an empty derived class of FileService.

    Although it doesn't do anything, it fixes an error with mypy:
        error: Value of type variable "FileModel" of "FileService" cannot be "JobScriptFile"
    """

    async def upsert(
        self,
        parent_id: int,
        filename: str,
        upload_content: str | bytes | AnyUrl | UploadFile | None,
        previous_filename: str | None = None,
        **upsert_kwargs,
    ) -> FileModel:
        """
        Upsert a file instance.
        """
        file_type: str = enforce_defined(
            upsert_kwargs.get("file_type", None),
            "File type must be defined when upserting a file.",
            raise_exc_class=ServiceError,
            raise_kwargs=dict(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY),
        )
        if file_type == FileType.ENTRYPOINT:
            await self.validate_entrypoint_file(parent_id, previous_filename or filename)
        return await super().upsert(parent_id, filename, upload_content, previous_filename, **upsert_kwargs)

    async def validate_entrypoint_file(self, parent_id: int, filename: str):
        """
        Validate that the entrypoint file is unique.
        """
        file_list = await self.find_children(parent_id)

        entry_point_names = {file.filename for file in file_list if file.file_type == FileType.ENTRYPOINT}

        no_entry_point = len(entry_point_names) == 0
        replacing_entry_point = filename in entry_point_names
        sanity_check = len(entry_point_names) <= 1

        require_condition(
            (no_entry_point or replacing_entry_point) and sanity_check,
            (
                "A job script can not have more than one entry point file. "
                "Consider deleting the existing one first. "
                "Found: {}".format(",".join(entry_point_names))
            ),
            raise_exc_class=ServiceError,
            raise_kwargs=dict(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY),
        )



import argparse
import asyncio
from contextlib import asynccontextmanager
from typing import AsyncIterator

from loguru import logger

from jobbergate_api.apps.dependencies import (
    Services,
    get_bucket_name,
    get_bucket_url,
    s3_bucket,
    service_factory,
)
from jobbergate_api.config import settings
from jobbergate_api.storage import engine_factory


@asynccontextmanager
async def cleanup_services(organization_id: str | None = None, commit=True) -> AsyncIterator[Services]:
    """Create a context manager for cleanup services."""
    override_base_name = organization_id if settings.MULTI_TENANCY_ENABLED else None
    bucket_name = get_bucket_name(override_base_name)
    s3_url = get_bucket_url()
    async with (
        s3_bucket(bucket_name, s3_url) as bucket,
        engine_factory.auto_session(override_base_name, commit=commit) as session,
    ):
        with service_factory(session, bucket) as services:
            yield services


async def run_cron_job(organization_id: str | None = None) -> None:
    """Run the cron jobs."""
    logger.info(f"Running cron jobs for organization ID: {organization_id}")
    async with cleanup_services(organization_id, commit=True) as services:
        for c in services.crud:
            await c.clean_unused_entries()

    async with cleanup_services(organization_id, commit=False) as services:
        for f in services.file:
            await f.clean_unused_files()

    logger.success(f"Finished running cron jobs for organization ID: {organization_id}")


async def main_async(targets: list[str] | list[None]) -> None:
    """Main function to run the cron jobs."""
    tasks = [run_cron_job(t) for t in targets]
    await asyncio.gather(*tasks)


def main() -> None:
    """Main function to run the cron jobs."""
    parser = argparse.ArgumentParser(description="Jobbergate API Cron Jobs")
    parser.add_argument(
        "--organization_id",
        "-i",
        type=str,
        action="append",
        default=[],
        help="Organization ID for multi-tenancy. Required when multi-tenancy is enabled.",
        required=settings.MULTI_TENANCY_ENABLED,
    )

    args = parser.parse_args()

    targets = args.organization_id if settings.MULTI_TENANCY_ENABLED else [None]
    asyncio.run(main_async(targets))



"""
JobSubmission resource schema.
"""

from typing import Self
from datetime import datetime
from collections.abc import Iterable

from pydantic import BaseModel, ConfigDict, Field, NonNegativeInt, field_validator

from jobbergate_api.apps.job_scripts.schemas import JobScriptBaseView, JobScriptDetailedView
from jobbergate_api.apps.job_submissions.constants import JobSubmissionStatus, SlurmJobState
from jobbergate_api.apps.schemas import LengthLimitedStr, TableResource
from jobbergate_api.meta_mapper import MetaField, MetaMapper

job_submission_meta_mapper = MetaMapper(
    id=MetaField(
        description="The unique database identifier for the instance",
        example=101,
    ),
    created_at=MetaField(
        description="The timestamp for when the instance was created",
        example="2023-08-18T13:55:37.172285",
    ),
    updated_at=MetaField(
        description="The timestamp for when the instance was last updated",
        example="2023-08-18T13:55:37.172285",
    ),
    name=MetaField(
        description="The unique name of the job submission",
        example="test-job-submission-77",
    ),
    description=MetaField(
        description="A text field providing a human-friendly description of the job_submission",
        example="Submission for the Foo job on sample 99 using the bar variant",
    ),
    owner_email=MetaField(
        description="The email of the owner/creator of the instance",
        example="tucker@omnivector.solutions",
    ),
    job_script_id=MetaField(
        description="The foreign-key to the job_script from which this instance was created",
        example=71,
    ),
    execution_directory=MetaField(
        description="The directory on the cluster where the job should be executed",
        example="/home/someuser/job-data/test-77",
    ),
    slurm_job_id=MetaField(
        description="The id for the slurm job executing this job_submission",
        example="1883",
    ),
    client_id=MetaField(
        description="The client_id of the cluster where this job submission should execute",
        example="mega-cluster-1",
    ),
    status=MetaField(
        description=f"The status of the job submission. Must be one of {JobSubmissionStatus.pretty_list()}",
        example=JobSubmissionStatus.CREATED,
    ),
    report_message=MetaField(
        description="The report message received from cluster-agent when a job submission is rejected",
        example="Unrecognized SBATCH arguments",
    ),
    slurm_job_state=MetaField(
        description="The Slurm Job state as reported by the agent.example",
        example="PENDING",
    ),
    slurm_job_info=MetaField(
        description="Detailed information about the Slurm Job as reported by the agent",
        example="""
            JobId=2 JobName=apptainer-test
            UserId=ubuntu(1000) GroupId=ubuntu(1000) MCS_label=N/A
            Priority=4294901758 Nice=0 Account=(null) QOS=normal
            JobState=RUNNING Reason=None Dependency=(null)
            Requeue=1 Restarts=0 BatchFlag=1 Reboot=0 ExitCode=0:0
            RunTime=00:01:57 TimeLimit=UNLIMITED TimeMin=N/A
            SubmitTime=2024-01-29T17:42:15 EligibleTime=2024-01-29T17:42:15
            AccrueTime=2024-01-29T17:42:15
            StartTime=2024-01-29T17:42:15 EndTime=Unknown Deadline=N/A
            SuspendTime=None SecsPreSuspend=0 LastSchedEval=2024-01-29T17:42:15 Scheduler=Main
            Partition=compute AllocNode:Sid=10.122.188.182:1278
            ReqNodeList=(null) ExcNodeList=(null)
            NodeList=democluster
            BatchHost=democluster
            NumNodes=1 NumCPUs=6 NumTasks=6 CPUs/Task=1 ReqB:S:C:T=0:0:*:*
            ReqTRES=cpu=6,mem=3903M,node=1,billing=6
            AllocTRES=cpu=6,node=1,billing=6
            Socks/Node=* NtasksPerN:B:S:C=0:0:*:* CoreSpec=*
            MinCPUsNode=1 MinMemoryNode=0 MinTmpDiskNode=0
            Features=(null) DelayBoot=00:00:00
            OverSubscribe=OK Contiguous=0 Licenses=(null) Network=(null)
            Command=(null)
            WorkDir=/tmp
            StdErr=/home/ubuntu/democluster/job.%J.err
            StdIn=/dev/null
            StdOut=/home/ubuntu/democluster/job.%J.out
            Power=
        """,
    ),
    is_archived=MetaField(
        description="Indicates if the job submission has been archived.",
        example=False,
    ),
    sbatch_arguments=MetaField(
        description="The arguments used to submit the job to the slurm queue",
        example=["--exclusive", "--job-name=example-job"],
    ),
    cloned_from_id=MetaField(
        description="Indicates the id this entry has been cloned from, if any.",
        example=101,
    ),
)


class JobSubmissionCreateRequest(BaseModel):
    """
    Request model for creating JobSubmission instances.
    """

    name: LengthLimitedStr
    description: LengthLimitedStr | None = None
    job_script_id: NonNegativeInt
    slurm_job_id: NonNegativeInt | None = None
    execution_directory: LengthLimitedStr | None = None
    client_id: LengthLimitedStr | None = None
    sbatch_arguments: list[LengthLimitedStr] | None = Field(None, max_length=50)

    @field_validator("execution_directory", mode="before")
    @classmethod
    def empty_str_to_none(cls, v):
        """Ensure empty strings are converted to None to avoid problems with Path downstream."""
        return v or None

    model_config = ConfigDict(json_schema_extra=job_submission_meta_mapper)


class JobSubmissionUpdateRequest(BaseModel):
    """
    Request model for updating JobSubmission instances.
    """

    name: LengthLimitedStr | None = None
    description: LengthLimitedStr | None = None
    execution_directory: LengthLimitedStr | None = None
    status: JobSubmissionStatus | None = None
    is_archived: bool | None = None

    @field_validator("execution_directory", mode="before")
    def empty_str_to_none(cls, v):
        """Ensure empty strings are converted to None to avoid problems with Path downstream."""
        return v or None

    model_config = ConfigDict(json_schema_extra=job_submission_meta_mapper)


class JobSubmissionBaseView(TableResource):
    """
    Base model to match the database for the JobSubmission resource.

    Omits parent relationship.
    """

    job_script_id: int | None = None
    slurm_job_id: int | None = None
    client_id: str
    status: JobSubmissionStatus
    slurm_job_state: SlurmJobState | None = None
    cloned_from_id: int | None = None

    model_config = ConfigDict(json_schema_extra=job_submission_meta_mapper)


class JobSubmissionListView(JobSubmissionBaseView):
    """
    Complete model to match the database for the JobSubmission resource in a list view.
    """

    job_script: JobScriptBaseView | None = None


class JobSubmissionDetailedView(JobSubmissionBaseView):
    """
    Complete model to match the database for the JobSubmission resource in a detailed view.
    """

    execution_directory: str | None
    report_message: str | None
    slurm_job_info: str | None
    sbatch_arguments: list[str] | None


class PendingJobSubmission(BaseModel):
    """
    Specialized model for the cluster-agent to pull pending job_submissions.

    Model also includes data from its job_script and application sources.
    """

    id: int
    name: str
    owner_email: str
    execution_directory: str | None = None
    execution_parameters: dict = Field(default_factory=dict)
    job_script: JobScriptDetailedView
    sbatch_arguments: list[str] | None = None

    model_config = ConfigDict(
        from_attributes=True, extra="ignore", json_schema_extra=job_submission_meta_mapper
    )


class ActiveJobSubmission(BaseModel):
    """
    Specialized model for the cluster-agent to pull an active job_submission.
    """

    id: int
    name: str
    slurm_job_id: int
    model_config = ConfigDict(from_attributes=True, extra="ignore")


class JobSubmissionAgentSubmittedRequest(BaseModel):
    """Request model for marking JobSubmission instances as SUBMITTED."""

    id: int
    slurm_job_id: NonNegativeInt
    slurm_job_state: SlurmJobState
    slurm_job_info: str
    slurm_job_state_reason: str | None = None

    model_config = ConfigDict(json_schema_extra=job_submission_meta_mapper)


class JobSubmissionAgentRejectedRequest(BaseModel):
    """Request model for marking JobSubmission instances as REJECTED."""

    id: int
    report_message: str

    model_config = ConfigDict(json_schema_extra=job_submission_meta_mapper)


class JobSubmissionAgentUpdateRequest(BaseModel):
    """Request model for updating JobSubmission instances."""

    slurm_job_id: NonNegativeInt
    slurm_job_state: SlurmJobState
    slurm_job_info: str
    slurm_job_state_reason: str | None = None

    model_config = ConfigDict(json_schema_extra=job_submission_meta_mapper)


class JobSubmissionAgentMaxTimes(BaseModel):
    """Model for the max_times field of the JobSubmissionMetricsMaxResponse."""

    max_time: int
    node_host: str
    step: int
    task: int

    model_config = ConfigDict(from_attributes=True, extra="ignore")


class JobSubmissionAgentMetricsRequest(BaseModel):
    """Request model for updating JobSubmission instances."""

    job_submission_id: int
    max_times: list[JobSubmissionAgentMaxTimes]


class JobSubmissionMetricSchema(BaseModel):
    """Model for the JobSubmissionMetric resource.

    Both `step` and `task` are optional fields, as they are not relevant when the metrics are
    queried over all the nodes. As well as, all measurements are both `int` and `float` due to
    the aggregation done by the time series database over time. For better understanding on the
    math behind the aggregation, please refer to the alembic revision ``99c3877d0f10``.
    """

    time: int | datetime
    node_host: str
    step: int | None = None
    task: int | None = None
    cpu_frequency: int | float
    cpu_time: float
    cpu_utilization: float
    gpu_memory: int | float
    gpu_utilization: float
    page_faults: int | float
    memory_rss: int | float
    memory_virtual: int | float
    disk_read: int | float
    disk_write: int | float

    model_config = ConfigDict(from_attributes=True, extra="ignore")

    @field_validator("time", mode="before")
    @classmethod
    def validate_time(cls, v: int | datetime) -> int:
        """Ensure that the time is always an int."""
        if isinstance(v, datetime):
            return int(v.timestamp())
        return v

    @classmethod
    def from_iterable(cls, iterable: Iterable, skip_optional: bool = False) -> Self:
        """Convert an iterable containing the fields of the model to an instance of the model."""
        if skip_optional:
            fields = list(field_name for field_name, field in cls.model_fields.items() if field.is_required())
        else:
            fields = list(cls.model_fields.keys())

        if len(fields) != len(list(iterable)):
            raise ValueError("The iterable must have the same length as the model fields.")

        return cls(**{field: value for field, value in zip(fields, iterable)})


class JobSubmissionMetricTimestamps(BaseModel):
    """Model for the timestamps of the JobSubmissionMetric resource."""

    min: int
    max: int

    model_config = ConfigDict(from_attributes=True, extra="ignore")


class JobProgressDetail(BaseModel):
    """
    Base model for the JobProgress resource.
    """

    id: int
    job_submission_id: int
    timestamp: datetime
    slurm_job_state: str | None = None
    additional_info: str | None = None
